from datetime import timedelta
from pathlib import Path

import yaml
from highway_dsl import (
    WorkflowBuilder,
)


def demonstrate_agentic_dev_platform_workflow():
    """
    Defines a massive, complex workflow for an AI agent that
    builds, tests, and deploys a full software platform using TDD.
    """

    builder = WorkflowBuilder("ai_agent_platform_builder_v1")

    # --- PHASE 1: SETUP & TEST GENERATION ---
    builder.task(
        "analyze_requirements",
        "llm.analyzer.ingest_specification",
        args=["{{project_spec_url}}"],
        result_key="project_plan",
    )
    builder.task(
        "setup_git_repository",
        "vcs.git.clone_and_setup_repo",
        args=["{{project_plan.repo_url}}"],
        result_key="workspace",
        dependencies=["analyze_requirements"],
    )

    # Parallel: Generate test suites for all microservices in the plan
    builder.foreach(
        "test_generation_loop",
        items="{{project_plan.microservices}}",
        loop_body=lambda fb: fb.task(
            "generate_test_stubs",
            "llm.test_generator.create_e2e_tests",
            args=["{{workspace}}", "{{item.service_name}}", "{{item.api_spec}}"],
        ).task(
            "commit_tests_to_git",
            "vcs.git.commit_files",
            args=[
                "{{workspace}}",
                "ci(tests): add initial test suite for {{item.service_name}}",
            ],
        ),
        dependencies=["setup_git_repository"],
    )

    # First test run: aggregate all failing tests to bootstrap the build loop
    builder.task(
        "run_initial_test_suite",
        "ci.run_all_tests",
        args=["{{workspace}}"],
        result_key="test_run_results",
        dependencies=["test_generation_loop"],
    )

    # --- PHASE 2: TDD BUILD & REFACTOR LOOP ---
    # This loop continues until all tests pass.
    builder.while_loop(
        "main_build_loop",
        condition="{{test_run_results.all_passed}} == false",
        loop_body=lambda wb: wb.task(
            "plan_next_coding_step",
            "llm.planner.analyze_failing_tests",
            args=["{{project_plan}}", "{{test_run_results.failing_tests}}"],
            result_key="coding_plan",  # e.g., {"action": "implement", "target": "user_api.create_user"}
        )
        .condition(
            "check_plan_action",
            condition="{{coding_plan.action}} == 'implement'",
            # IF "implement": Write new code
            if_true=lambda ib: ib.task(
                "implement_code",
                "llm.coder.write_implementation",
                args=[
                    "{{workspace}}",
                    "{{coding_plan.target}}",
                    "{{coding_plan.prompt}}",
                ],
                result_key="code_diff",
            ),
            # ELSE "refactor": Refactor existing code
            if_false=lambda rb: rb.task(
                "refactor_code",
                "llm.refactor.rewrite_code",
                args=[
                    "{{workspace}}",
                    "{{coding_plan.target}}",
                    "{{coding_plan.reason}}",
                ],
                result_key="code_diff",
            ),
        )
        .task(
            "apply_and_commit_code",
            "vcs.git.apply_and_commit",
            args=["{{workspace}}", "{{code_diff}}", "feat: {{coding_plan.target}}"],
        )
        # Re-run all tests in parallel to get the next loop's condition
        .parallel(
            "run_full_test_suite",
            branches={
                "unit": lambda b: b.task(
                    "run_unit_tests",
                    "ci.run_unit_tests",
                    args=["{{workspace}}"],
                    result_key="unit_results",
                ),
                "integration": lambda b: b.task(
                    "run_integration_tests",
                    "ci.run_integration_tests",
                    args=["{{workspace}}"],
                    result_key="int_results",
                ),
                "e2e": lambda b: b.task(
                    "run_e2e_tests",
                    "ci.run_e2e_tests",
                    args=["{{workspace}}"],
                    result_key="e2e_results",
                ),
            },
        )
        .task(
            "aggregate_test_results",
            "ci.aggregate_results",
            args=["{{unit_results}}", "{{int_results}}", "{{e2e_results}}"],
            result_key="test_run_results",  # This updates the loop condition variable
        ),
        dependencies=["run_initial_test_suite"],
    )

    # --- PHASE 3: DEPLOYMENT ---
    # This phase only starts after the main_build_loop is complete.
    builder.task(
        "provision_infrastructure",
        "iac.terraform.apply",
        args=["./terraform"],
        result_key="infra_outputs",
        dependencies=["main_build_loop"],
    )

    # Deploy all microservices in parallel using Ansible
    builder.parallel(
        "run_ansible_deployments",
        branches={
            "api": lambda b: b.task(
                "ansible_deploy_api",
                "deploy.ansible.run_playbook",
                args=["playbooks/api.yml"],
            ),
            "frontend": lambda b: b.task(
                "ansible_deploy_frontend",
                "deploy.ansible.run_playbook",
                args=["playbooks/frontend.yml"],
            ),
            "database": lambda b: b.task(
                "ansible_deploy_db",
                "deploy.ansible.run_playbook",
                args=["playbooks/database.yml"],
            ),
            "workers": lambda b: b.task(
                "ansible_deploy_workers",
                "deploy.ansible.run_playbook",
                args=["playbooks/workers.yml"],
            ),
        },
        dependencies=["provision_infrastructure"],
    )

    # Wait for services to come online
    builder.wait(
        "wait_for_services_health",
        timedelta(minutes=5),
        dependencies=["run_ansible_deployments"],
    )

    # --- PHASE 4: OBSERVABILITY & FINAL VERIFICATION ---
    # Setup monitoring and logging in parallel
    builder.parallel(
        "setup_observability",
        branches={
            "metrics": lambda b: b.task(
                "configure_prometheus",
                "observe.prometheus.apply_config",
                args=["{{infra_outputs.service_endpoints}}"],
            ),
            "logging": lambda b: b.task(
                "configure_loki",
                "observe.loki.apply_config",
                args=["{{infra_outputs.log_streams}}"],
            ),
            "tracing": lambda b: b.task(
                "configure_tempo",
                "observe.tempo.apply_config",
            ),
        },
        dependencies=["wait_for_services_health"],
    )

    # Final check: Run smoke tests AND verify observability data is flowing
    builder.task(
        "run_production_smoke_tests",
        "ci.run_smoke_tests",
        args=["{{infra_outputs.app_url}}"],
        result_key="smoke_test_results",
        dependencies=["setup_observability"],
    )
    builder.task(
        "verify_observability_data",
        "observe.prometheus.query_metrics",
        args=["{{project_plan.key_metrics}}"],
        result_key="metrics_check",
        dependencies=["run_production_smoke_tests"],
    )

    # Final step: report success
    builder.task(
        "notify_user_on_slack",
        "notify.slack.send_message",
        args=["Platform build and deployment complete. Metrics are active."],
        dependencies=["verify_observability_data"],
    )

    # Build the final workflow object
    workflow = builder.build()
    workflow.set_variables(
        {
            "project_spec_url": "http://s3.com/specs/my_project.md",
            "ci_runner_pool": "default-pool",
        },
    )

    return workflow


def extract_yaml_content(content):
    """Extract only the YAML portion from the output file"""
    lines = content.split("\n")

    # The YAML content starts after the header line and ends before the footer
    # Header: "--- AI AGENTIC DEVELOPER WORKFLOW YAML (MASSIVE) ---" at line 0
    # YAML starts from line 1 (description: '')
    # Footer starts with the separator line and continues with success message

    # Find the first line that contains YAML content
    yaml_start = 0
    for i, line in enumerate(lines):
        if line.strip().startswith("description:") or line.strip().startswith("name:"):
            yaml_start = i
            break

    # Find the end - look for the separator line before the footer
    yaml_end = len(lines)
    for i in range(len(lines) - 1, yaml_start, -1):
        line = lines[i].strip()
        if line == "---------------------------------------------------":
            # This is typically the start of the footer, so everything before this is YAML
            yaml_end = i
            break
        if "Successfully generated" in line:
            # Sometimes the separator might be missing, so look for the success message
            yaml_end = i
            # Then go backwards to find the actual end of YAML content
            while yaml_end > yaml_start and (
                lines[yaml_end - 1].strip() == ""
                or lines[yaml_end - 1].strip().startswith("---")
                or lines[yaml_end - 1].strip()
                == "---------------------------------------------------"
            ):
                yaml_end -= 1
            break

    # Extract and return the YAML content
    yaml_content = "\n".join(lines[yaml_start:yaml_end])
    return yaml_content.strip()


def test_agentic_ai_software_workflow():
    """Test the agentic AI software workflow generates expected YAML"""
    workflow = demonstrate_agentic_dev_platform_workflow()
    generated_yaml = workflow.to_yaml()
    generated_data = yaml.safe_load(generated_yaml)

    # Load expected output
    expected_file = (
        Path(__file__).parent / "data" / "test_driven_agentic_ai_software_workflow.yaml"
    )
    with open(expected_file) as f:
        content = f.read()
        expected_content = extract_yaml_content(content)
        expected_data = yaml.safe_load(expected_content)

    # Compare the structure and key elements
    assert generated_data["name"] == expected_data["name"]
    assert len(generated_data["tasks"]) == len(expected_data["tasks"])

    # Check that certain key tasks exist
    assert "analyze_requirements" in generated_data["tasks"]
    assert "test_generation_loop" in generated_data["tasks"]
    assert "main_build_loop" in generated_data["tasks"]
    assert "provision_infrastructure" in generated_data["tasks"]
    assert "run_ansible_deployments" in generated_data["tasks"]
    assert "notify_user_on_slack" in generated_data["tasks"]

    # Check operator types
    assert generated_data["tasks"]["test_generation_loop"]["operator_type"] == "foreach"
    assert generated_data["tasks"]["main_build_loop"]["operator_type"] == "while"
    assert (
        generated_data["tasks"]["run_ansible_deployments"]["operator_type"]
        == "parallel"
    )
    assert generated_data["tasks"]["analyze_requirements"]["operator_type"] == "task"

    # Validate the generated YAML matches expected (ignoring the header)
    assert generated_data["name"] == expected_data["name"]
    assert generated_data["version"] == expected_data["version"]
    assert set(generated_data["tasks"].keys()) == set(expected_data["tasks"].keys())


if __name__ == "__main__":
    test_agentic_ai_software_workflow()



"""
Test suite to validate all the workflow examples work as expected
"""

from tests.test_agentic_ai_software_workflow import test_agentic_ai_software_workflow
from tests.test_car_factory_workflow import test_car_factory_workflow
from tests.test_car_factory_workflow_with_fluent_builder import (
    test_car_factory_workflow_with_fluent_builder,
)
from tests.test_complex_agentic_workflow import test_complex_agentic_workflow
from tests.test_example_usage import test_example_usage_workflows


def test_all_workflow_examples():
    """Run all workflow example tests"""
    test_car_factory_workflow()
    test_car_factory_workflow_with_fluent_builder()
    test_complex_agentic_workflow()
    test_example_usage_workflows()
    test_agentic_ai_software_workflow()


if __name__ == "__main__":
    test_all_workflow_examples()



from datetime import timedelta
from pathlib import Path

import yaml
from highway_dsl import (
    RetryPolicy,
    TimeoutPolicy,
    WorkflowBuilder,
)


def create_vehicle_sub_workflow_builder(builder: WorkflowBuilder) -> WorkflowBuilder:
    """
    Defines the sub-workflow for building a single vehicle.
    This workflow is built fluently using the new builder features.
    """

    # 1. Get detailed specs for the specific vehicle
    builder.task(
        "check_vehicle_specs",
        "factory.mes.get_specs_for_vin",
        args=["{{item.vin}}"],  # 'item' is the context from the ForEach loop
        result_key="specs",
    )

    # 2. Conditional: Is this an EV or an ICE car? (Uses new fluent builder)
    builder.condition(
        "route_by_drivetrain",
        condition="{{specs.drivetrain_type}} == 'EV'",
        if_true=lambda b: b.task(
            "build_ev_drivetrain",
            "factory.assembly.build_battery_and_motor",
            args=["{{specs.battery_sku}}", "{{specs.motor_sku}}"],
            result_key="drivetrain_assembly",
        ),
        if_false=lambda b: b.task(
            "build_ice_drivetrain",
            "factory.assembly.build_engine_and_transmission",
            args=["{{specs.engine_sku}}", "{{specs.transmission_sku}}"],
            result_key="drivetrain_assembly",
        ),
    )

    # 3. Parallel Assembly (Uses new fluent builder)
    # This automatically depends on the 'route_by_drivetrain' conditional.
    builder.parallel(
        "parallel_main_assembly",
        branches={
            "frame_build": lambda b: b.task(
                "weld_frame",
                "factory.robotics.weld_chassis",
                args=["{{specs.frame_model}}"],
            ).task("prime_frame", "factory.robotics.apply_primer"),
            "electronics_build": lambda b: b.task(
                "assemble_electronics_harness",
                "factory.assembly.build_electronics",
                args=["{{specs.electronics_package}}"],
            ),
            "drivetrain_prep": lambda b: b.task(
                "mount_drivetrain_subassembly",
                "factory.assembly.prep_drivetrain_for_marriage",
                args=["{{drivetrain_assembly}}"],
            ),
        },
    )

    # 4. Synchronization (Fan-In): "Marriage" of parts
    # This task automatically depends on all branches of the parallel step.
    builder.task(
        "final_assembly",
        "factory.assembly.final_marriage",
        args=["{{item.vin}}"],
        dependencies=[
            "prime_frame",
            "assemble_electronics_harness",
            "mount_drivetrain_subassembly",
        ],
        result_key="assembled_vehicle",
    )

    # 5. Paint and Initial QA
    builder.task(
        "paint_vehicle",
        "factory.robotics.paint_body",
        args=["{{assembled_vehicle}}", "{{specs.color}}"],
    )
    builder.task(
        "begin_qa_inspection",
        "human.qa.inspect_vehicle",
        args=["{{item.vin}}"],
        result_key="qa_report",  # e.g., {"passed": false, "issues": [...]}
        dependencies=["paint_vehicle"],
        timeout_policy=TimeoutPolicy(timeout=timedelta(hours=2), kill_on_timeout=True),
    )

    # 6. QA Rework (Uses new WhileLoop operator)
    # This replaces the complex, limited conditional logic from before.
    builder.while_loop(
        "qa_rework_loop",
        condition="{{qa_report.passed}} == false",
        loop_body=lambda b: b.task(
            "perform_rework",
            "human.rework.fix_issues",
            args=["{{item.vin}}", "{{qa_report.issues}}"],
            result_key="rework_report",
        ).task(
            "re_inspect_vehicle",
            "human.qa.inspect_rework",
            args=["{{item.vin}}", "{{rework_report}}"],
            result_key="qa_report",  # Critically, this updates the loop variable
        ),
    )

    # 7. Mark as complete (depends on the while_loop finishing)
    builder.task(
        "mark_vehicle_complete",
        "factory.mes.update_status",
        args=["{{item.vin}}", "BUILT"],
        dependencies=["qa_rework_loop"],
    )

    return builder


def demonstrate_car_factory_workflow():
    """
    Defines the main car factory workflow, integrating the
    vehicle sub-workflow with the ForEach operator.
    """

    # 1. Define the main workflow builder
    main_builder = WorkflowBuilder("car_factory_build_v2")
    main_builder.task(
        "get_build_manifest",
        "factory.erp.get_daily_build_manifest",
        result_key="manifest",
        retry_policy=RetryPolicy(max_retries=5, delay=timedelta(minutes=1)),
    )

    # 2. Use the fluent ForEach operator
    main_builder.foreach(
        "build_vehicle_loop",
        items="{{manifest.vehicles}}",
        loop_body=lambda b: create_vehicle_sub_workflow_builder(b),
    )

    # 5. Add final aggregation steps
    #    These depend on the *entire loop* finishing.
    main_builder.task(
        "generate_shipping_labels",
        "factory.logistics.print_labels_for_manifest",
        args=["{{manifest.id}}", "{{build_vehicle_loop_results}}"],
        result_key="shipping_labels",
        dependencies=["build_vehicle_loop"],
    )
    main_builder.task(
        "notify_logistics",
        "factory.erp.notify_logistics_of_completion",
        args=["{{manifest.id}}"],
        dependencies=["generate_shipping_labels"],
    )

    workflow = main_builder.build()
    workflow.set_variables(
        {"erp_api_key": "secret_abc_123", "mes_endpoint": "http://10.0.0.5/api"},
    )

    return workflow


def extract_yaml_content(content):
    """Extract only the YAML portion from the output file"""
    lines = content.split("\n")

    # The YAML content starts after the header line and ends before the footer
    # Header: "--- CAR FACTORY WORKFLOW YAML (FLUENT + WHILE LOOP) ---" at line 0
    # YAML starts from line 1 (description: '')
    # Footer starts with the separator line and continues with success message

    # Find the first line that contains YAML content
    yaml_start = 0
    for i, line in enumerate(lines):
        if line.strip().startswith("description:") or line.strip().startswith("name:"):
            yaml_start = i
            break

    # Find the end - look for the separator line before the footer
    yaml_end = len(lines)
    for i in range(len(lines) - 1, yaml_start, -1):
        line = lines[i].strip()
        if line == "---------------------------------------------":
            # This is typically the start of the footer, so everything before this is YAML
            yaml_end = i
            break
        if "Successfully generated" in line:
            # Sometimes the separator might be missing, so look for the success message
            yaml_end = i
            # Then go backwards to find the actual end of YAML content
            while yaml_end > yaml_start and (
                lines[yaml_end - 1].strip() == ""
                or lines[yaml_end - 1].strip().startswith("---")
                or lines[yaml_end - 1].strip()
                == "---------------------------------------------"
            ):
                yaml_end -= 1
            break

    # Extract and return the YAML content
    yaml_content = "\n".join(lines[yaml_start:yaml_end])
    return yaml_content.strip()


def test_car_factory_workflow_with_fluent_builder():
    """Test the car factory workflow with fluent builder generates expected YAML"""
    workflow = demonstrate_car_factory_workflow()
    generated_yaml = workflow.to_yaml()
    generated_data = yaml.safe_load(generated_yaml)

    # Load expected output
    expected_file = (
        Path(__file__).parent / "data" / "car_factory_workflow_with_fluent_builder.yaml"
    )
    with open(expected_file) as f:
        content = f.read()
        expected_content = extract_yaml_content(content)
        expected_data = yaml.safe_load(expected_content)

    # Compare the structure and key elements
    assert generated_data["name"] == expected_data["name"]
    assert len(generated_data["tasks"]) == len(expected_data["tasks"])

    # Check that certain key tasks exist
    assert "get_build_manifest" in generated_data["tasks"]
    assert "build_vehicle_loop" in generated_data["tasks"]
    assert "check_vehicle_specs" in generated_data["tasks"]
    assert "route_by_drivetrain" in generated_data["tasks"]
    assert "mark_vehicle_complete" in generated_data["tasks"]
    assert generated_data["tasks"]["get_build_manifest"]["operator_type"] == "task"
    assert generated_data["tasks"]["build_vehicle_loop"]["operator_type"] == "foreach"
    assert (
        generated_data["tasks"]["route_by_drivetrain"]["operator_type"] == "condition"
    )

    # Validate the generated YAML matches expected (ignoring the header)
    assert generated_data["name"] == expected_data["name"]
    assert generated_data["version"] == expected_data["version"]
    assert set(generated_data["tasks"].keys()) == set(expected_data["tasks"].keys())


if __name__ == "__main__":
    test_car_factory_workflow_with_fluent_builder()



from datetime import timedelta
from pathlib import Path

import yaml
from highway_dsl import (
    RetryPolicy,
    Workflow,
    WorkflowBuilder,
)


def demonstrate_while_loop():
    """Demonstrate the while loop operator"""
    builder = WorkflowBuilder("qa_rework_workflow")

    builder.task("start_qa", "workflows.tasks.start_qa", result_key="qa_results")

    builder.while_loop(
        "qa_rework_loop",
        condition="{{qa_results.status}} == 'failed'",
        loop_body=lambda b: b.task(
            "perform_rework", "workflows.tasks.perform_rework",
        ).task("re_run_qa", "workflows.tasks.run_qa", result_key="qa_results"),
    )

    builder.task(
        "finalize_product",
        "workflows.tasks.finalize_product",
        dependencies=["qa_rework_loop"],
    )

    workflow = builder.build()

    workflow.set_variables({"product_id": "product-123"})

    return workflow


def create_complex_workflow() -> Workflow:
    """Create a complex workflow using the Python DSL"""

    builder = WorkflowBuilder("data_processing_pipeline")

    builder.task("start", "workflows.tasks.initialize", result_key="init_data")
    builder.task(
        "validate",
        "workflows.tasks.validate_data",
        args=["{{init_data}}"],
        result_key="validated_data",
    )

    builder.condition(
        "check_quality",
        condition="{{validated_data.quality_score}} > 0.8",
        if_true=lambda b: b.task(
            "high_quality_processing",
            "workflows.tasks.advanced_processing",
            args=["{{validated_data}}"],
            retry_policy=RetryPolicy(
                max_retries=5, delay=timedelta(seconds=10), backoff_factor=2.0,
            ),
        ),
        if_false=lambda b: b.task(
            "standard_processing",
            "workflows.tasks.basic_processing",
            args=["{{validated_data}}"],
        ),
    )

    builder.parallel(
        "parallel_processing",
        branches={
            "branch_a": lambda b: b.task(
                "transform_a", "workflows.tasks.transform_a", result_key="transformed_a",
            ).task(
                "enrich_a",
                "workflows.tasks.enrich_data",
                args=["{{transformed_a}}"],
                result_key="enriched_a",
            ),
            "branch_b": lambda b: b.task(
                "transform_b", "workflows.tasks.transform_b", result_key="transformed_b",
            ).task(
                "enrich_b",
                "workflows.tasks.enrich_data",
                args=["{{transformed_b}}"],
                result_key="enriched_b",
            ),
        },
        dependencies=["high_quality_processing", "standard_processing"],
    )

    builder.task(
        "aggregate",
        "workflows.tasks.aggregate_results",
        dependencies=["enrich_a", "enrich_b"],
        result_key="final_result",
    )
    builder.wait("wait_notification", timedelta(hours=1))
    builder.task(
        "notify", "workflows.tasks.send_notification", args=["{{final_result}}"],
    )

    workflow = builder.build()

    workflow.set_variables(
        {
            "environment": "production",
            "batch_size": 1000,
            "notify_email": "team@company.com",
        },
    )

    return workflow


def demonstrate_basic_workflow():
    """Show a simple complete workflow using just the builder"""

    workflow = (
        WorkflowBuilder("simple_etl")
        .task("extract", "etl.extract_data", result_key="raw_data")
        .task(
            "transform",
            "etl.transform_data",
            args=["{{raw_data}}"],
            result_key="transformed_data",
        )
        .retry(max_retries=3, delay=timedelta(seconds=10))
        .task("load", "etl.load_data", args=["{{transformed_data}}"])
        .timeout(timeout=timedelta(minutes=30))
        .wait("wait_next", timedelta(hours=24))
        .task("cleanup", "etl.cleanup")
        .build()
    )

    workflow.set_variables(
        {"database_url": "postgresql://localhost/mydb", "chunk_size": 1000},
    )

    return workflow


def extract_yaml_content(content):
    """Extract only the YAML portion from the output file"""
    lines = content.split("\n")

    # For example_usage, the file contains output from multiple workflows
    # The first YAML workflow is in the "Converting to YAML" section
    # Find the line with "=== Converting to YAML ==="
    yaml_header_idx = -1
    for i, line in enumerate(lines):
        if line.strip() == "=== Converting to YAML ===":
            yaml_header_idx = i
            break

    if yaml_header_idx != -1:
        # The YAML content starts after the header
        yaml_start = yaml_header_idx + 1

        # Find the end - either the next section header or the next '==='
        yaml_end = len(lines)
        for i in range(yaml_start, len(lines)):
            if lines[i].strip().startswith("===") and i > yaml_start:
                yaml_end = i
                break

        # Extract and return the YAML content
        yaml_content = "\n".join(lines[yaml_start:yaml_end])
        return yaml_content.strip()

    # Fallback: look for the first YAML content as before
    yaml_start = 0
    for i, line in enumerate(lines):
        if line.strip().startswith("description:") or line.strip().startswith("name:"):
            yaml_start = i
            break

    # Find the end by looking for the next section or success message
    yaml_end = len(lines)
    for i in range(len(lines) - 1, yaml_start, -1):
        line = lines[i].strip()
        if line.startswith("==="):
            yaml_end = i
            # Find the actual end of YAML content by going backwards
            while yaml_end > yaml_start and lines[yaml_end - 1].strip() == "":
                yaml_end -= 1
            break
        if "Successfully generated" in line:
            yaml_end = i
            while yaml_end > yaml_start and (
                lines[yaml_end - 1].strip() == ""
                or lines[yaml_end - 1].strip().startswith("---")
            ):
                yaml_end -= 1
            break

    # Extract and return the YAML content
    yaml_content = "\n".join(lines[yaml_start:yaml_end])
    return yaml_content.strip()


def test_example_usage_workflows():
    """Test the example usage workflows generate expected YAML"""
    # Test complex workflow (the first one in the combined output)
    complex_workflow = create_complex_workflow()
    complex_yaml = complex_workflow.to_yaml()
    complex_data = yaml.safe_load(complex_yaml)

    # Load expected output
    expected_file = Path(__file__).parent / "data" / "example_usage.yaml"
    with open(expected_file) as f:
        content = f.read()
        # The example_usage.py output contains multiple workflows combined,
        # so we need to extract the complex workflow portion specifically
        expected_content = extract_yaml_content(content)
        expected_data = yaml.safe_load(expected_content)

    # Compare basic properties (the first workflow in the file should be the complex one)
    # The content contains multiple workflows, so let's just check that we have a complex structure
    if "name" in expected_data and expected_data["name"] == "data_processing_pipeline":
        # This is the complex workflow
        assert complex_data["name"] == expected_data["name"]
        assert complex_data["version"] == expected_data["version"]
    # If the first workflow isn't the complex one, we just verify the structure exists
    assert "name" in complex_data
    assert "tasks" in complex_data
    assert isinstance(complex_data["tasks"], dict)

    # Test while loop workflow
    while_workflow = demonstrate_while_loop()
    while_yaml = while_workflow.to_yaml()
    while_data = yaml.safe_load(while_yaml)

    assert while_data["name"] == "qa_rework_workflow"
    assert "qa_rework_loop" in while_data["tasks"]
    assert while_data["tasks"]["qa_rework_loop"]["operator_type"] == "while"

    # Test basic workflow
    basic_workflow = demonstrate_basic_workflow()
    basic_yaml = basic_workflow.to_yaml()
    basic_data = yaml.safe_load(basic_yaml)

    assert basic_data["name"] == "simple_etl"
    assert "extract" in basic_data["tasks"]
    assert "transform" in basic_data["tasks"]
    assert "load" in basic_data["tasks"]
    assert basic_data["tasks"]["transform"]["operator_type"] == "task"

    # Check that complex workflow has multiple types of operators
    task_count = 0
    condition_count = 0
    parallel_count = 0
    for task in complex_data["tasks"].values():
        if task["operator_type"] == "task":
            task_count += 1
        elif task["operator_type"] == "condition":
            condition_count += 1
        elif task["operator_type"] == "parallel":
            parallel_count += 1

    assert task_count > 0
    assert condition_count > 0
    assert parallel_count > 0


if __name__ == "__main__":
    test_example_usage_workflows()



from datetime import timedelta
from pathlib import Path

import yaml
from highway_dsl import (
    ConditionOperator,
    ForEachOperator,
    ParallelOperator,
    RetryPolicy,
    TaskOperator,
    TimeoutPolicy,
    WorkflowBuilder,
)


def demonstrate_car_factory_workflow():
    """
    Defines an extremely complex workflow for a car factory,
    pushing the limits of the DSL's graph capabilities.
    """

    # 1. Start with the builder for the initial task
    workflow = (
        WorkflowBuilder("car_factory_build_v1")
        .task(
            "get_build_manifest",
            "factory.erp.get_daily_build_manifest",
            result_key="manifest",
            retry_policy=RetryPolicy(max_retries=5, delay=timedelta(minutes=1)),
        )
        .build()
    )

    # 2. Add the main ForEach loop to iterate over each vehicle in the manifest
    # The 'task_chain' defines the sub-workflow for *one* vehicle.
    workflow.add_task(
        ForEachOperator(
            task_id="build_vehicle_loop",
            items="{{manifest.vehicles}}",  # e.g., [{"vin": "...", "model": "...", "type": "EV"}, ...]
            task_chain=["check_vehicle_specs"],  # Entry point for the sub-workflow
            dependencies=["get_build_manifest"],
        ),
    )

    # --- Define the "Build One Vehicle" Sub-Workflow ---
    # All tasks from here until the final 'builder.task' are part
    # of the ForEach 'task_chain', executed for each 'item'.

    # 3. Get detailed specs for the specific vehicle
    workflow.add_task(
        TaskOperator(
            task_id="check_vehicle_specs",
            function="factory.mes.get_specs_for_vin",
            args=["{{item.vin}}"],
            result_key="specs",
            # This task has no dependencies *within the chain*
            # as it is the entry point.
        ),
    )

    # 4. Conditional: Is this an EV or an ICE car?
    workflow.add_task(
        ConditionOperator(
            task_id="route_by_drivetrain",
            condition="{{specs.drivetrain_type}} == 'EV'",
            if_true="build_ev_drivetrain",
            if_false="build_ice_drivetrain",
            dependencies=["check_vehicle_specs"],
        ),
    )

    # --- Drivetrain Branches (mutually exclusive) ---
    workflow.add_task(
        TaskOperator(
            task_id="build_ev_drivetrain",
            function="factory.assembly.build_battery_and_motor",
            args=["{{specs.battery_sku}}", "{{specs.motor_sku}}"],
            result_key="drivetrain_assembly",
            dependencies=["route_by_drivetrain"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="build_ice_drivetrain",
            function="factory.assembly.build_engine_and_transmission",
            args=["{{specs.engine_sku}}", "{{specs.transmission_sku}}"],
            result_key="drivetrain_assembly",
            dependencies=["route_by_drivetrain"],
        ),
    )

    # 5. Parallel Assembly: Build frame, electronics, and run diagnostics
    #    This is a "fan-in-fan-out". It waits for *either* drivetrain
    #    branch to finish, then fans out to 3 parallel tracks.
    workflow.add_task(
        ParallelOperator(
            task_id="parallel_main_assembly",
            branches={
                "frame_build": ["weld_frame", "prime_frame"],
                "electronics_build": ["assemble_electronics_harness"],
                "drivetrain_prep": ["mount_drivetrain_subassembly"],
            },
            # This task depends on *both* conditional outcomes.
            # Since only one will run, the dependency logic in the
            # engine must be "run if at least one dependency is met"
            # (or, more simply, it depends on the *drivetrain_assembly*
            # result, which is output by both tasks).
            # For simplicity, we'll depend on the tasks themselves.
            dependencies=["build_ev_drivetrain", "build_ice_drivetrain"],
        ),
    )

    # --- Tasks inside the Parallel Branches ---
    workflow.add_task(
        TaskOperator(
            task_id="weld_frame",
            function="factory.robotics.weld_chassis",
            args=["{{specs.frame_model}}"],
            dependencies=["parallel_main_assembly"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="prime_frame",
            function="factory.robotics.apply_primer",
            dependencies=["weld_frame"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="assemble_electronics_harness",
            function="factory.assembly.build_electronics",
            args=["{{specs.electronics_package}}"],
            dependencies=["parallel_main_assembly"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="mount_drivetrain_subassembly",
            function="factory.assembly.prep_drivetrain_for_marriage",
            args=["{{drivetrain_assembly}}"],
            dependencies=["parallel_main_assembly"],
        ),
    )

    # 6. Synchronization (Fan-In): "Marriage" of parts
    #    Waits for all parallel branches to complete.
    workflow.add_task(
        TaskOperator(
            task_id="final_assembly",
            function="factory.assembly.final_marriage",
            args=["{{item.vin}}"],
            dependencies=[
                "prime_frame",  # End of frame_build branch
                "assemble_electronics_harness",  # End of electronics_build
                "mount_drivetrain_subassembly",  # End of drivetrain_prep
            ],
            result_key="assembled_vehicle",
        ),
    )

    # 7. Paint the fully assembled body
    workflow.add_task(
        TaskOperator(
            task_id="paint_vehicle",
            function="factory.robotics.paint_body",
            args=["{{assembled_vehicle}}", "{{specs.color}}"],
            dependencies=["final_assembly"],
        ),
    )

    # 8. Human-in-the-Loop Quality Assurance (with timeout)
    workflow.add_task(
        TaskOperator(
            task_id="begin_qa_inspection",
            function="human.qa.inspect_vehicle",
            args=["{{item.vin}}"],
            result_key="qa_report",
            dependencies=["paint_vehicle"],
            timeout_policy=TimeoutPolicy(timeout=timedelta(hours=2)),
        ),
    )

    # 9. QA Conditional: Did it pass?
    workflow.add_task(
        ConditionOperator(
            task_id="check_qa_results",
            condition="{{qa_report.passed}}",
            if_true="mark_vehicle_complete",  # End of this loop
            if_false="perform_rework",  # Go to rework sub-path
            dependencies=["begin_qa_inspection"],
        ),
    )

    # --- Rework Sub-Path (if QA fails) ---
    workflow.add_task(
        TaskOperator(
            task_id="perform_rework",
            function="human.rework.fix_issues",
            args=["{{item.vin}}", "{{qa_report.issues}}"],
            result_key="rework_report",
            dependencies=["check_qa_results"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="final_rework_qa",
            function="human.qa.inspect_rework",
            args=["{{item.vin}}", "{{rework_report}}"],
            result_key="final_qa_status",
            dependencies=["perform_rework"],
        ),
    )

    # 10. Final check. This demonstrates the "fixed rework attempts"
    #     limitation of a DAG.
    workflow.add_task(
        ConditionOperator(
            task_id="check_final_qa",
            condition="{{final_qa_status.passed}}",
            if_true="mark_vehicle_complete",  # Pass: End of this loop
            if_false="flag_for_manual_intervention",  # Fail: End of this loop
            dependencies=["final_rework_qa"],
        ),
    )

    # --- Loop "Sink" (End) Tasks ---
    # These tasks are the defined "ends" of the sub-workflow.
    # The ForEach operator knows an iteration is done when one of these is hit.
    workflow.add_task(
        TaskOperator(
            task_id="mark_vehicle_complete",
            function="factory.mes.update_status",
            args=["{{item.vin}}", "BUILT"],
            dependencies=["check_qa_results", "check_final_qa"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="flag_for_manual_intervention",
            function="factory.mes.update_status",
            args=["{{item.vin}}", "QA_FAIL_MANUAL_HOLD"],
            dependencies=["check_final_qa"],
        ),
    )

    # --- End of "Build One Vehicle" Sub-Workflow ---

    # 11. Final Aggregation Step (after the ForEach loop)
    # Re-attach the builder to add a final step that depends
    # on the *entire loop* finishing.
    builder = WorkflowBuilder(workflow.name, existing_workflow=workflow)
    builder._current_task = "build_vehicle_loop"  # Depends on the loop itself

    workflow = (
        builder.task(
            "generate_shipping_labels",
            "factory.logistics.print_labels_for_manifest",
            args=["{{manifest.id}}", "{{build_vehicle_loop_results}}"],
            result_key="shipping_labels",
        )
        .task(
            "notify_logistics",
            "factory.erp.notify_logistics_of_completion",
            args=["{{manifest.id}}"],
        )
        .build()
    )

    workflow.set_variables(
        {"erp_api_key": "secret_abc_123", "mes_endpoint": "http://10.0.0.5/api"},
    )

    return workflow


def extract_yaml_content(content):
    """Extract only the YAML portion from the output file"""
    lines = content.split("\n")

    # The YAML content starts after the header line and ends before the footer
    # Header: "--- CAR FACTORY WORKFLOW YAML (COMPLEX) ---" at line 0
    # YAML starts from line 1 (description: '')
    # Footer starts with the separator line and continues with success message

    # Find the first line that contains YAML content
    yaml_start = 0
    for i, line in enumerate(lines):
        if line.strip().startswith("description:") or line.strip().startswith("name:"):
            yaml_start = i
            break

    # Find the end - look for the separator line before the footer
    yaml_end = len(lines)
    for i in range(len(lines) - 1, yaml_start, -1):
        line = lines[i].strip()
        if line == "---------------------------------------------":
            # This is typically the start of the footer, so everything before this is YAML
            yaml_end = i
            break
        if "Successfully generated" in line:
            # Sometimes the separator might be missing, so look for the success message
            yaml_end = i
            # Then go backwards to find the actual end of YAML content
            while yaml_end > yaml_start and (
                lines[yaml_end - 1].strip() == ""
                or lines[yaml_end - 1].strip().startswith("---")
                or lines[yaml_end - 1].strip()
                == "---------------------------------------------"
            ):
                yaml_end -= 1
            break

    # Extract and return the YAML content
    yaml_content = "\n".join(lines[yaml_start:yaml_end])
    return yaml_content.strip()


def test_car_factory_workflow():
    """Test the car factory workflow generates expected YAML"""
    workflow = demonstrate_car_factory_workflow()
    generated_yaml = workflow.to_yaml()
    generated_data = yaml.safe_load(generated_yaml)

    # Load expected output
    expected_file = Path(__file__).parent / "data" / "car_factory_workflow.yaml"
    with open(expected_file) as f:
        content = f.read()
        expected_content = extract_yaml_content(content)
        expected_data = yaml.safe_load(expected_content)

    # Compare the structure and key elements
    assert generated_data["name"] == expected_data["name"]
    assert len(generated_data["tasks"]) == len(expected_data["tasks"])

    # Check that certain key tasks exist
    assert "get_build_manifest" in generated_data["tasks"]
    assert "build_vehicle_loop" in generated_data["tasks"]
    assert "mark_vehicle_complete" in generated_data["tasks"]
    assert generated_data["tasks"]["get_build_manifest"]["operator_type"] == "task"
    assert generated_data["tasks"]["build_vehicle_loop"]["operator_type"] == "foreach"

    # Validate the generated YAML matches expected (ignoring the header)
    assert generated_data["name"] == expected_data["name"]
    assert generated_data["version"] == expected_data["version"]
    assert set(generated_data["tasks"].keys()) == set(expected_data["tasks"].keys())


if __name__ == "__main__":
    test_car_factory_workflow()



import os  # noqa: F401
from pathlib import Path
import pytest
from highway_dsl import Workflow, TaskOperator, ConditionOperator, ParallelOperator, ForEachOperator, WhileOperator

# Define the output directory for Mermaid diagrams
MERMAID_TEST_RESULTS_DIR = Path("/tmp/highway_dsl/mermaid_test_results")

@pytest.fixture(scope="session", autouse=True)
def create_mermaid_test_dir():
    """Create the directory for Mermaid test results if it doesn't exist."""
    MERMAID_TEST_RESULTS_DIR.mkdir(parents=True, exist_ok=True)

def save_mermaid_file(workflow_name: str, mermaid_content: str):
    """Saves the Mermaid diagram to a file."""
    file_path = MERMAID_TEST_RESULTS_DIR / f"{workflow_name}.mermaid"
    file_path.write_text(mermaid_content)

def test_single_task_workflow_to_mermaid():
    # Create a workflow with a single task
    workflow = Workflow(name="single_task_workflow", start_task="task1")
    workflow.add_task(TaskOperator(task_id="task1", function="func1"))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    [*] --> task1
    task1 --> [*]"""

    # Compare the generated diagram with the expected one
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))

def test_simple_linear_workflow_to_mermaid():
    # Create a simple linear workflow
    workflow = Workflow(name="simple_linear_workflow", start_task="task1")
    workflow.add_task(TaskOperator(task_id="task1", function="func1"))
    workflow.add_task(TaskOperator(task_id="task2", function="func2", dependencies=["task1"]))
    workflow.add_task(TaskOperator(task_id="task3", function="func3", dependencies=["task2"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    task1 --> task2
    task2 --> task3
    [*] --> task1
    task3 --> [*]"""

    # Compare the generated diagram with the expected one
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))

def test_conditional_workflow_to_mermaid():
    # Create a conditional workflow
    workflow = Workflow(name="conditional_workflow", start_task="start_task")
    workflow.add_task(TaskOperator(task_id="start_task", function="start_func"))
    workflow.add_task(ConditionOperator(task_id="condition", condition="x > 10", if_true="true_task", if_false="false_task", dependencies=["start_task"]))
    workflow.add_task(TaskOperator(task_id="true_task", function="true_func"))
    workflow.add_task(TaskOperator(task_id="false_task", function="false_func"))
    workflow.add_task(TaskOperator(task_id="end_task", function="end_func", dependencies=["true_task", "false_task"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    [*] --> start_task
    start_task --> condition
    condition --> true_task : True
    condition --> false_task : False
    true_task --> end_task
    false_task --> end_task
    end_task --> [*]"""

    # Compare the generated diagram with the expected one
    # Normalize line endings and remove leading/trailing whitespace
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))


def test_parallel_workflow_to_mermaid():
    # Create a parallel workflow
    workflow = Workflow(name="parallel_workflow", start_task="start_task")
    workflow.add_task(TaskOperator(task_id="start_task", function="start_func"))
    workflow.add_task(ParallelOperator(task_id="parallel_task", branches={"b1": ["t1"], "b2": ["t2"]}, dependencies=["start_task"]))
    workflow.add_task(TaskOperator(task_id="t1", function="t1_func", dependencies=["parallel_task"]))
    workflow.add_task(TaskOperator(task_id="t2", function="t2_func", dependencies=["parallel_task"]))
    workflow.add_task(TaskOperator(task_id="end_task", function="end_func", dependencies=["t1", "t2"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    [*] --> start_task
    start_task --> parallel_task
    state parallel_task {
        state "Branch 1" as b1
        --
        state "Branch 2" as b2
    }
    parallel_task --> t1
    parallel_task --> t2
    t1 --> end_task
    t2 --> end_task
    end_task --> [*]"""

    # Compare the generated diagram with the expected one
    # Normalize line endings and remove leading/trailing whitespace
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))

def test_while_workflow_with_composite_state_to_mermaid():
    # Create a while workflow
    workflow = Workflow(name="while_workflow_with_composite_state", start_task="start_task")
    workflow.add_task(TaskOperator(task_id="start_task", function="start_func"))
    workflow.add_task(WhileOperator(task_id="while_task", condition="x > 10", loop_body=[TaskOperator(task_id="t1", function="t1_func", description="Task 1")], dependencies=["start_task"]))
    workflow.add_task(TaskOperator(task_id="end_task", function="end_func", dependencies=["while_task"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    [*] --> start_task
    start_task --> while_task
    state while_task {
        state "Task 1" as t1
    }
    while_task --> end_task
    end_task --> [*]"""

    # Compare the generated diagram with the expected one
    # Normalize line endings and remove leading/trailing whitespace
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))

def test_foreach_workflow_with_composite_state_to_mermaid():
    # Create a foreach workflow
    workflow = Workflow(name="foreach_workflow_with_composite_state", start_task="start_task")
    workflow.add_task(TaskOperator(task_id="start_task", function="start_func"))
    workflow.add_task(ForEachOperator(task_id="foreach_task", items="items", loop_body=[TaskOperator(task_id="t1", function="t1_func", description="Task 1")], dependencies=["start_task"]))
    workflow.add_task(TaskOperator(task_id="end_task", function="end_func", dependencies=["foreach_task"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    [*] --> start_task
    start_task --> foreach_task
    state foreach_task {
        state "Task 1" as t1
    }
    foreach_task --> end_task
    end_task --> [*]"""

    # Compare the generated diagram with the expected one
    # Normalize line endings and remove leading/trailing whitespace
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))

def test_parallel_workflow_with_composite_state_to_mermaid():
    # Create a parallel workflow
    workflow = Workflow(name="parallel_workflow_with_composite_state", start_task="start_task")
    workflow.add_task(TaskOperator(task_id="start_task", function="start_func"))
    workflow.add_task(ParallelOperator(task_id="parallel_task", branches={"b1": ["t1"], "b2": ["t2"]}, dependencies=["start_task"]))
    workflow.add_task(TaskOperator(task_id="t1", function="t1_func", dependencies=["parallel_task"]))
    workflow.add_task(TaskOperator(task_id="t2", function="t2_func", dependencies=["parallel_task"]))
    workflow.add_task(TaskOperator(task_id="end_task", function="end_func", dependencies=["t1", "t2"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    [*] --> start_task
    start_task --> parallel_task
    state parallel_task {
        state "Branch 1" as b1
        --
        state "Branch 2" as b2
    }
    parallel_task --> t1
    parallel_task --> t2
    t1 --> end_task
    t2 --> end_task
    end_task --> [*]"""

    # Compare the generated diagram with the expected one
    # Normalize line endings and remove leading/trailing whitespace
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))

def test_state_with_description_to_mermaid():
    # Create a workflow with a state that has a description
    workflow = Workflow(name="state_with_description_workflow", start_task="task1")
    workflow.add_task(TaskOperator(task_id="task1", function="func1", description="This is a task with a description"))
    workflow.add_task(TaskOperator(task_id="task2", function="func2", dependencies=["task1"]))

    # Generate the Mermaid diagram
    mermaid_diagram = workflow.to_mermaid()
    save_mermaid_file(workflow.name, mermaid_diagram)

    # Define the expected Mermaid diagram
    expected_mermaid = """stateDiagram-v2
    state "This is a task with a description" as task1
    [*] --> task1
    task1 --> task2
    task2 --> [*]"""

    # Compare the generated diagram with the expected one
    assert "\n".join(sorted(mermaid_diagram.strip().split("\n"))) == "\n".join(sorted(expected_mermaid.strip().split("\n")))


from datetime import timedelta
from pathlib import Path

import yaml
from highway_dsl import (
    ConditionOperator,
    ForEachOperator,
    ParallelOperator,
    TaskOperator,
    TimeoutPolicy,
    WorkflowBuilder,
)


def demonstrate_complex_agentic_workflow():
    """
    Defines a complex agentic workflow using the DSL's
    correct syntax, including manual graph construction.
    """

    # 1. Start with the builder for the first step
    workflow = (
        WorkflowBuilder("customer_support_agent_v2")
        .task(
            "get_pending_tasks",
            "agent.get_pending_tasks",
            result_key="task_list_obj",
        )
        .build()
    )

    # 2. Add the ForEach operator to loop over the tasks.
    # This requires defining the "task_chain" (sub-workflow)
    # that will be executed for each item.
    # The task_chain IDs are: "analyze", "check_review", "route_by_review"
    workflow.add_task(
        ForEachOperator(
            task_id="process_all_tasks",
            items="{{task_list_obj.tasks}}",
            task_chain=["analyze", "check_review", "route_by_review"],
            dependencies=["get_pending_tasks"],
        ),
    )

    # --- Define the "process_one_task" sub-workflow ---
    # These tasks are part of the ForEach task_chain.
    # Note: The DSL does not seem to have a way to pass the
    # loop item ("task") to the sub-tasks automatically,
    # so we assume the execution engine handles this context.

    workflow.add_task(
        TaskOperator(
            task_id="analyze",
            function="agent.analyze_sentiment",
            args=["{{item.text}}"],  # Assuming 'item' is the loop variable
            result_key="sentiment_analysis",
            # No dependencies, as it's the start of the task_chain
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="check_review",
            function="agent.needs_human_review",
            args=["{{sentiment_analysis.sentiment}}"],
            result_key="review_check",
            dependencies=["analyze"],
        ),
    )
    # 3. Add the conditional (if/else) operator
    workflow.add_task(
        ConditionOperator(
            task_id="route_by_review",
            condition="{{review_check.review_needed}}",
            if_true="human_review_branch_start",  # ID of task to run if true
            if_false="auto_approve_branch_start",  # ID of task to run if false
            dependencies=["check_review"],
        ),
    )

    # --- Define the "auto-approve" branch ---
    workflow.add_task(
        TaskOperator(
            task_id="auto_approve_branch_start",
            function="agent.generate_response",
            args=["{{item.text}}"],
            result_key="auto_response",
            dependencies=["route_by_review"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="send_auto_response",
            function="agent.send_final_response",
            args=["{{item.id}}", "{{auto_response.response}}"],
            dependencies=["auto_approve_branch_start"],
        ),
    )

    # --- Define the "human review" branch (with parallelism) ---

    # 4. Add the Parallel operator
    workflow.add_task(
        ParallelOperator(
            task_id="human_review_branch_start",
            # Defines the parallel branches by listing the task IDs
            # that form the *end* of each branch.
            branches={
                "branch_draft": ["generate_draft"],
                "branch_docs": ["find_docs"],
            },
            dependencies=["route_by_review"],
        ),
    )
    # Define tasks *inside* the parallel branches
    workflow.add_task(
        TaskOperator(
            task_id="generate_draft",
            function="agent.generate_response",
            args=["{{item.text}}"],
            result_key="draft_response",
            dependencies=["human_review_branch_start"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="find_docs",
            function="agent.find_documentation",
            args=["{{item.text}}"],
            result_key="docs",
            dependencies=["human_review_branch_start"],
        ),
    )

    # 5. Fan-in: A task that waits for *both* parallel branches
    workflow.add_task(
        TaskOperator(
            task_id="request_approval",
            function="human.request_human_approval",
            args=["{{item}}", "{{draft_response.response}}", "{{docs.doc_url}}"],
            result_key="approval",
            dependencies=["generate_draft", "find_docs"],  # Waits for both
            timeout_policy=TimeoutPolicy(timeout=timedelta(hours=4)),
        ),
    )

    # 6. Another conditional based on human approval
    workflow.add_task(
        ConditionOperator(
            task_id="process_approval",
            condition="{{approval.approved}}",
            if_true="send_human_approved_response",
            if_false="archive_rejected_task",
            dependencies=["request_approval"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="send_human_approved_response",
            function="agent.send_final_response",
            args=["{{item.id}}", "{{approval.final_response}}"],
            dependencies=["process_approval"],
        ),
    )
    workflow.add_task(
        TaskOperator(
            task_id="archive_rejected_task",
            function="agent.archive_task",
            args=["{{item.id}}", "Rejected by human review"],
            dependencies=["process_approval"],
        ),
    )

    # 7. Final aggregation task (fan-in for the ForEach loop)
    # We re-attach the builder to fluently add the last step.
    builder = WorkflowBuilder(workflow.name, existing_workflow=workflow)

    # Manually set the builder's last task to the ForEach operator,
    # so the next task depends on it.
    builder._current_task = "process_all_tasks"

    workflow = builder.task(
        "create_summary_report",
        "agent.create_summary_report",
        # Assumes the ForEach operator aggregates results
        args=["{{process_all_tasks_results}}"],
        result_key="final_report",
    ).build()

    workflow.set_variables({"api_key": "xyz-123", "default_user": "agent_bot"})

    return workflow


def extract_yaml_content(content):
    """Extract only the YAML portion from the output file"""
    lines = content.split("\n")

    # The YAML content starts after the header line and ends before the footer
    # Header: "--- COMPLEX AGENTIC WORKFLOW YAML (CORRECTED) ---" at line 0
    # YAML starts from line 1 (description: '')
    # Footer starts with the separator line and continues with success message

    # Find the first line that contains YAML content
    yaml_start = 0
    for i, line in enumerate(lines):
        if line.strip().startswith("description:") or line.strip().startswith("name:"):
            yaml_start = i
            break

    # Find the end - look for the separator line before the footer
    yaml_end = len(lines)
    for i in range(len(lines) - 1, yaml_start, -1):
        line = lines[i].strip()
        if line == "-------------------------------------------------":
            # This is typically the start of the footer, so everything before this is YAML
            yaml_end = i
            break
        if "Successfully generated" in line:
            # Sometimes the separator might be missing, so look for the success message
            yaml_end = i
            # Then go backwards to find the actual end of YAML content
            while yaml_end > yaml_start and (
                lines[yaml_end - 1].strip() == ""
                or lines[yaml_end - 1].strip().startswith("---")
                or lines[yaml_end - 1].strip()
                == "-------------------------------------------------"
            ):
                yaml_end -= 1
            break

    # Extract and return the YAML content
    yaml_content = "\n".join(lines[yaml_start:yaml_end])
    return yaml_content.strip()


def test_complex_agentic_workflow():
    """Test the complex agentic workflow generates expected YAML"""
    workflow = demonstrate_complex_agentic_workflow()
    generated_yaml = workflow.to_yaml()
    generated_data = yaml.safe_load(generated_yaml)

    # Load expected output
    expected_file = Path(__file__).parent / "data" / "complex_agentic_workflow.yaml"
    with open(expected_file) as f:
        content = f.read()
        expected_content = extract_yaml_content(content)
        expected_data = yaml.safe_load(expected_content)

    # Compare the structure and key elements
    assert generated_data["name"] == expected_data["name"]
    assert len(generated_data["tasks"]) == len(expected_data["tasks"])

    # Check that certain key tasks exist
    assert "get_pending_tasks" in generated_data["tasks"]
    assert "process_all_tasks" in generated_data["tasks"]
    assert "analyze" in generated_data["tasks"]
    assert "route_by_review" in generated_data["tasks"]
    assert "create_summary_report" in generated_data["tasks"]
    assert generated_data["tasks"]["get_pending_tasks"]["operator_type"] == "task"
    assert generated_data["tasks"]["process_all_tasks"]["operator_type"] == "foreach"
    assert generated_data["tasks"]["route_by_review"]["operator_type"] == "condition"
    assert (
        generated_data["tasks"]["human_review_branch_start"]["operator_type"]
        == "parallel"
    )

    # Validate the generated YAML matches expected (ignoring the header)
    assert generated_data["name"] == expected_data["name"]
    assert generated_data["version"] == expected_data["version"]
    assert set(generated_data["tasks"].keys()) == set(expected_data["tasks"].keys())


if __name__ == "__main__":
    test_complex_agentic_workflow()



# workflow_dsl.py
from abc import ABC
from collections.abc import Callable
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Optional, Union

import yaml
from pydantic import BaseModel, ConfigDict, Field, model_validator


class OperatorType(Enum):
    TASK = "task"
    CONDITION = "condition"
    WAIT = "wait"
    PARALLEL = "parallel"
    FOREACH = "foreach"
    SWITCH = "switch"
    TRY_CATCH = "try_catch"
    WHILE = "while"
    EMIT_EVENT = "emit_event"
    WAIT_FOR_EVENT = "wait_for_event"


class RetryPolicy(BaseModel):
    max_retries: int = Field(3, description="Maximum number of retries")
    delay: timedelta = Field(timedelta(seconds=5), description="Delay between retries")
    backoff_factor: float = Field(2.0, description="Factor by which to increase delay")


class TimeoutPolicy(BaseModel):
    timeout: timedelta = Field(..., description="Timeout duration")
    kill_on_timeout: bool = Field(
        True, description="Whether to kill the task on timeout",
    )


class BaseOperator(BaseModel, ABC):
    task_id: str
    operator_type: OperatorType
    dependencies: list[str] = Field(default_factory=list)
    retry_policy: RetryPolicy | None = None
    timeout_policy: TimeoutPolicy | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)
    description: str = Field(default="", description="Task description")
    # Phase 3: Callback hooks
    on_success_task_id: str | None = Field(None, description="Task to run on success")
    on_failure_task_id: str | None = Field(None, description="Task to run on failure")
    is_internal_loop_task: bool = Field(
        default=False, exclude=True,
    )  # Mark if task is internal to a loop

    model_config = ConfigDict(use_enum_values=True, arbitrary_types_allowed=True)


class TaskOperator(BaseOperator):
    function: str
    args: list[Any] = Field(default_factory=list)
    kwargs: dict[str, Any] = Field(default_factory=dict)
    result_key: str | None = None
    operator_type: OperatorType = Field(OperatorType.TASK, frozen=True)


class ConditionOperator(BaseOperator):
    condition: str
    if_true: str | None
    if_false: str | None
    operator_type: OperatorType = Field(OperatorType.CONDITION, frozen=True)


class WaitOperator(BaseOperator):
    wait_for: timedelta | datetime | str
    operator_type: OperatorType = Field(OperatorType.WAIT, frozen=True)

    @model_validator(mode="before")
    @classmethod
    def parse_wait_for(cls, data: Any) -> Any:
        if isinstance(data, dict) and "wait_for" in data:
            wait_for = data["wait_for"]
            if isinstance(wait_for, str):
                if wait_for.startswith("duration:"):
                    data["wait_for"] = timedelta(seconds=float(wait_for.split(":")[1]))
                elif wait_for.startswith("datetime:"):
                    data["wait_for"] = datetime.fromisoformat(wait_for.split(":", 1)[1])
        return data

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        data = super().model_dump(**kwargs)
        wait_for = data["wait_for"]
        if isinstance(wait_for, timedelta):
            data["wait_for"] = f"duration:{wait_for.total_seconds()}"
        elif isinstance(wait_for, datetime):
            data["wait_for"] = f"datetime:{wait_for.isoformat()}"
        return data


class ParallelOperator(BaseOperator):
    branches: dict[str, list[str]] = Field(default_factory=dict)
    timeout: int | None = Field(None, description="Optional timeout in seconds for branch execution")
    operator_type: OperatorType = Field(OperatorType.PARALLEL, frozen=True)


class ForEachOperator(BaseOperator):
    items: str
    loop_body: list[
        Union[
            TaskOperator,
            ConditionOperator,
            WaitOperator,
            ParallelOperator,
            "ForEachOperator",
            "WhileOperator",
            "EmitEventOperator",
            "WaitForEventOperator",
            "SwitchOperator",
        ]
    ] = Field(default_factory=list)
    operator_type: OperatorType = Field(OperatorType.FOREACH, frozen=True)


class WhileOperator(BaseOperator):
    condition: str
    loop_body: list[
        Union[
            TaskOperator,
            ConditionOperator,
            WaitOperator,
            ParallelOperator,
            ForEachOperator,
            "WhileOperator",
            "EmitEventOperator",
            "WaitForEventOperator",
            "SwitchOperator",
        ]
    ] = Field(default_factory=list)
    operator_type: OperatorType = Field(OperatorType.WHILE, frozen=True)


class EmitEventOperator(BaseOperator):
    """Phase 2: Emit an event that other workflows can wait for."""
    event_name: str = Field(..., description="Name of the event to emit")
    payload: dict[str, Any] = Field(default_factory=dict, description="Event payload data")
    operator_type: OperatorType = Field(OperatorType.EMIT_EVENT, frozen=True)


class WaitForEventOperator(BaseOperator):
    """Phase 2: Wait for an external event with optional timeout."""
    event_name: str = Field(..., description="Name of the event to wait for")
    timeout_seconds: int | None = Field(None, description="Timeout in seconds (None = wait forever)")
    operator_type: OperatorType = Field(OperatorType.WAIT_FOR_EVENT, frozen=True)


class SwitchOperator(BaseOperator):
    """Phase 4: Multi-branch switch/case operator."""
    switch_on: str = Field(..., description="Expression to evaluate for switch")
    cases: dict[str, str] = Field(default_factory=dict, description="Map of case values to task IDs")
    default: str | None = Field(None, description="Default task ID if no case matches")
    operator_type: OperatorType = Field(OperatorType.SWITCH, frozen=True)


class Workflow(BaseModel):
    name: str
    version: str = "1.1.0"
    description: str = ""
    tasks: dict[
        str,
        TaskOperator | ConditionOperator | WaitOperator | ParallelOperator | ForEachOperator | WhileOperator | EmitEventOperator | WaitForEventOperator | SwitchOperator,
    ] = Field(default_factory=dict)
    variables: dict[str, Any] = Field(default_factory=dict)
    start_task: str | None = None

    # Phase 1: Scheduling metadata
    schedule: str | None = Field(None, description="Cron expression for scheduled execution")
    start_date: datetime | None = Field(None, description="When the schedule becomes active")
    catchup: bool = Field(False, description="Whether to backfill missed runs")
    is_paused: bool = Field(False, description="Whether the workflow is paused")
    tags: list[str] = Field(default_factory=list, description="Workflow categorization tags")
    max_active_runs: int = Field(1, description="Maximum number of concurrent runs")
    default_retry_policy: RetryPolicy | None = Field(None, description="Default retry policy for all tasks")

    @model_validator(mode="before")
    @classmethod
    def validate_workflow_name_and_version(cls, data: Any) -> Any:
        """Validate workflow name and version don't contain '__' (double underscore).

        The double underscore is reserved as a separator for display purposes:
        {workflow_name}__{version}__{step_name}

        Workflow names must match: ^[a-z][a-z0-9_]*$ (lowercase, alphanumeric, single underscore)
        Workflow versions must match: ^[a-zA-Z0-9._-]+$ (semver compatible)
        """
        import re

        if isinstance(data, dict):
            name = data.get("name", "")
            version = data.get("version", "")

            # Check for double underscore (reserved separator)
            if "__" in name:
                msg = f"Workflow name '{name}' cannot contain '__' (double underscore) - it's reserved as a separator"
                raise ValueError(msg)

            if "__" in version:
                msg = f"Workflow version '{version}' cannot contain '__' (double underscore) - it's reserved as a separator"
                raise ValueError(msg)

            # Validate workflow name format
            if name and not re.match(r"^[a-z][a-z0-9_]*$", name):
                msg = f"Workflow name '{name}' must start with lowercase letter and contain only lowercase letters, digits, and single underscores"
                raise ValueError(msg)

            # Validate workflow version format (semver compatible)
            if version and not re.match(r"^[a-zA-Z0-9._-]+$", version):
                msg = f"Workflow version '{version}' must contain only alphanumeric characters, dots, hyphens, and underscores (semver compatible)"
                raise ValueError(msg)

        return data

    @model_validator(mode="before")
    @classmethod
    def validate_tasks(cls, data: Any) -> Any:
        if isinstance(data, dict) and "tasks" in data:
            validated_tasks = {}
            operator_classes: dict[str, type[BaseOperator]] = {
                OperatorType.TASK.value: TaskOperator,
                OperatorType.CONDITION.value: ConditionOperator,
                OperatorType.WAIT.value: WaitOperator,
                OperatorType.PARALLEL.value: ParallelOperator,
                OperatorType.FOREACH.value: ForEachOperator,
                OperatorType.WHILE.value: WhileOperator,
                OperatorType.EMIT_EVENT.value: EmitEventOperator,
                OperatorType.WAIT_FOR_EVENT.value: WaitForEventOperator,
                OperatorType.SWITCH.value: SwitchOperator,
            }
            for task_id, task_data in data["tasks"].items():
                operator_type = task_data.get("operator_type")
                if operator_type and operator_type in operator_classes:
                    operator_class = operator_classes[operator_type]
                    validated_tasks[task_id] = operator_class.model_validate(task_data)
                else:
                    msg = f"Unknown operator type: {operator_type}"
                    raise ValueError(msg)
            data["tasks"] = validated_tasks
        return data

    def add_task(
        self,
        task: TaskOperator | ConditionOperator | WaitOperator | ParallelOperator | ForEachOperator | WhileOperator | EmitEventOperator | WaitForEventOperator | SwitchOperator,
    ) -> "Workflow":
        self.tasks[task.task_id] = task
        return self

    def set_variables(self, variables: dict[str, Any]) -> "Workflow":
        self.variables.update(variables)
        return self

    def set_start_task(self, task_id: str) -> "Workflow":
        self.start_task = task_id
        return self

    # Phase 1: Scheduling methods
    def set_schedule(self, cron: str) -> "Workflow":
        """Set the cron schedule for this workflow."""
        self.schedule = cron
        return self

    def set_start_date(self, start_date: datetime) -> "Workflow":
        """Set when the schedule becomes active."""
        self.start_date = start_date
        return self

    def set_catchup(self, enabled: bool) -> "Workflow":
        """Set whether to backfill missed runs."""
        self.catchup = enabled
        return self

    def set_paused(self, paused: bool) -> "Workflow":
        """Set whether the workflow is paused."""
        self.is_paused = paused
        return self

    def add_tags(self, *tags: str) -> "Workflow":
        """Add tags to the workflow."""
        self.tags.extend(tags)
        return self

    def set_max_active_runs(self, count: int) -> "Workflow":
        """Set maximum number of concurrent runs."""
        self.max_active_runs = count
        return self

    def set_default_retry_policy(self, policy: RetryPolicy) -> "Workflow":
        """Set default retry policy for all tasks."""
        self.default_retry_policy = policy
        return self

    def to_yaml(self) -> str:
        data = self.model_dump(mode="json", by_alias=True, exclude_none=True)
        return yaml.dump(data, default_flow_style=False)

    def to_json(self) -> str:
        return self.model_dump_json(indent=2)

    def to_mermaid(self) -> str:
        """ convert to mermaid state diagram format """
        lines = ["stateDiagram-v2"]
        
        all_dependencies = {dep for task in self.tasks.values() for dep in task.dependencies}
        
        for task_id, task in self.tasks.items():
            # Add state with description for regular tasks
            if task.description and not isinstance(task, (ForEachOperator, WhileOperator)):
                lines.append(f'    state "{task.description}" as {task_id}')
            
            # Add dependencies
            if not task.dependencies:
                if self.start_task == task_id or not self.start_task:
                    lines.append(f'    [*] --> {task_id}')
            else:
                for dep in task.dependencies:
                    lines.append(f'    {dep} --> {task_id}')
            
            # Add transitions for conditional operator
            if isinstance(task, ConditionOperator):
                if task.if_true:
                    lines.append(f'    {task_id} --> {task.if_true} : True')
                if task.if_false:
                    lines.append(f'    {task_id} --> {task.if_false} : False')

            # Add composite state for parallel operator
            if isinstance(task, ParallelOperator):
                lines.append(f'    state {task_id} {{')
                for i, branch in enumerate(task.branches):
                    lines.append(f'        state "Branch {i+1}" as {branch}')
                    if i < len(task.branches) - 1:
                        lines.append('        --')
                lines.append('    }')

            # Add composite state for foreach operator
            if isinstance(task, ForEachOperator):
                lines.append(f'    state {task_id} {{')
                for sub_task in task.loop_body:
                    if sub_task.description:
                        lines.append(f'        state "{sub_task.description}" as {sub_task.task_id}')
                    else:
                        lines.append(f'        {sub_task.task_id}')
                lines.append('    }')

            # Add composite state for while operator
            if isinstance(task, WhileOperator):
                lines.append(f'    state {task_id} {{')
                for sub_task in task.loop_body:
                    if sub_task.description:
                        lines.append(f'        state "{sub_task.description}" as {sub_task.task_id}')
                    else:
                        lines.append(f'        {sub_task.task_id}')
                lines.append('    }')

            # End states
            if task_id not in all_dependencies:
                if not (isinstance(task, ConditionOperator) and (task.if_true or task.if_false)):
                    lines.append(f'    {task_id} --> [*]')
                
        return "\n".join(lines)

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "Workflow":
        data = yaml.safe_load(yaml_str)
        return cls.model_validate(data)

    @classmethod
    def from_json(cls, json_str: str) -> "Workflow":
        return cls.model_validate_json(json_str)


class WorkflowBuilder:
    def __init__(
        self,
        name: str,
        existing_workflow: Workflow | None = None,
        parent: Optional["WorkflowBuilder"] = None,
    ) -> None:
        if existing_workflow:
            self.workflow = existing_workflow
        else:
            self.workflow = Workflow(
                name=name,
                version="1.1.0",
                description="",
                tasks={},
                variables={},
                start_task=None,
                schedule=None,
                start_date=None,
                catchup=False,
                is_paused=False,
                tags=[],
                max_active_runs=1,
                default_retry_policy=None,
            )
        self._current_task: str | None = None
        self.parent = parent

    def _add_task(
        self,
        task: TaskOperator | ConditionOperator | WaitOperator | ParallelOperator | ForEachOperator | WhileOperator | EmitEventOperator | WaitForEventOperator | SwitchOperator,
        **kwargs: Any,
    ) -> None:
        dependencies = kwargs.get("dependencies", [])
        if self._current_task and not dependencies:
            dependencies.append(self._current_task)

        task.dependencies = sorted(set(dependencies))

        self.workflow.add_task(task)
        self._current_task = task.task_id

    def task(self, task_id: str, function: str, **kwargs: Any) -> "WorkflowBuilder":
        task = TaskOperator(task_id=task_id, function=function, **kwargs)
        self._add_task(task, **kwargs)
        return self

    def condition(
        self,
        task_id: str,
        condition: str,
        if_true: Callable[["WorkflowBuilder"], "WorkflowBuilder"],
        if_false: Callable[["WorkflowBuilder"], "WorkflowBuilder"],
        **kwargs: Any,
    ) -> "WorkflowBuilder":
        true_builder = if_true(WorkflowBuilder(f"{task_id}_true", parent=self))
        false_builder = if_false(WorkflowBuilder(f"{task_id}_false", parent=self))

        true_tasks = list(true_builder.workflow.tasks.keys())
        false_tasks = list(false_builder.workflow.tasks.keys())

        task = ConditionOperator(
            task_id=task_id,
            condition=condition,
            if_true=true_tasks[0] if true_tasks else None,
            if_false=false_tasks[0] if false_tasks else None,
            **kwargs,
        )

        self._add_task(task, **kwargs)

        for task_obj in true_builder.workflow.tasks.values():
            # Only add the condition task as dependency, preserve original dependencies
            if task_id not in task_obj.dependencies:
                task_obj.dependencies.append(task_id)
            self.workflow.add_task(task_obj)
        for task_obj in false_builder.workflow.tasks.values():
            # Only add the condition task as dependency, preserve original dependencies
            if task_id not in task_obj.dependencies:
                task_obj.dependencies.append(task_id)
            self.workflow.add_task(task_obj)

        self._current_task = task_id
        return self

    def wait(
        self, task_id: str, wait_for: timedelta | datetime | str, **kwargs: Any,
    ) -> "WorkflowBuilder":
        task = WaitOperator(task_id=task_id, wait_for=wait_for, **kwargs)
        self._add_task(task, **kwargs)
        return self

    def parallel(
        self,
        task_id: str,
        branches: dict[str, Callable[["WorkflowBuilder"], "WorkflowBuilder"]],
        **kwargs: Any,
    ) -> "WorkflowBuilder":
        branch_builders = {}
        for name, branch_func in branches.items():
            branch_builder = branch_func(
                WorkflowBuilder(f"{task_id}_{name}", parent=self),
            )
            branch_builders[name] = branch_builder

        branch_tasks = {
            name: list(builder.workflow.tasks.keys())
            for name, builder in branch_builders.items()
        }

        task = ParallelOperator(task_id=task_id, branches=branch_tasks, **kwargs)

        self._add_task(task, **kwargs)

        for builder in branch_builders.values():
            for task_obj in builder.workflow.tasks.values():
                # Only add the parallel task as dependency to non-internal tasks,
                # preserve original dependencies
                if (
                    not getattr(task_obj, "is_internal_loop_task", False)
                    and task_id not in task_obj.dependencies
                ):
                    task_obj.dependencies.append(task_id)
                self.workflow.add_task(task_obj)

        self._current_task = task_id
        return self

    def foreach(
        self,
        task_id: str,
        items: str,
        loop_body: Callable[["WorkflowBuilder"], "WorkflowBuilder"],
        **kwargs: Any,
    ) -> "WorkflowBuilder":
        # Create a temporary builder for the loop body.
        temp_builder = WorkflowBuilder(f"{task_id}_loop", parent=self)
        loop_builder = loop_body(temp_builder)
        loop_tasks = list(loop_builder.workflow.tasks.values())

        # Mark all loop body tasks as internal to prevent parallel dependency injection
        for task_obj in loop_tasks:
            task_obj.is_internal_loop_task = True

        # Create the foreach operator
        task = ForEachOperator(
            task_id=task_id,
            items=items,
            loop_body=loop_tasks,
            **kwargs,
        )

        # Add the foreach task to workflow to establish initial dependencies
        self._add_task(task, **kwargs)

        # Add the foreach task as dependency to the FIRST task in the loop body
        # and preserve the original dependency chain within the loop
        if loop_tasks:
            first_task = loop_tasks[0]
            if task_id not in first_task.dependencies:
                first_task.dependencies.append(task_id)

            # Add all loop tasks to workflow
            for task_obj in loop_tasks:
                self.workflow.add_task(task_obj)

        self._current_task = task_id
        return self

    def while_loop(
        self,
        task_id: str,
        condition: str,
        loop_body: Callable[["WorkflowBuilder"], "WorkflowBuilder"],
        **kwargs: Any,
    ) -> "WorkflowBuilder":
        loop_builder = loop_body(WorkflowBuilder(f"{task_id}_loop", parent=self))
        loop_tasks = list(loop_builder.workflow.tasks.values())

        # Mark all loop body tasks as internal to prevent parallel dependency injection
        for task_obj in loop_tasks:
            task_obj.is_internal_loop_task = True

        task = WhileOperator(
            task_id=task_id,
            condition=condition,
            loop_body=loop_tasks,
            **kwargs,
        )

        self._add_task(task, **kwargs)

        # Fix: Only add the while task as dependency to the FIRST task in the loop body
        # and preserve the original dependency chain within the loop
        if loop_tasks:
            first_task = loop_tasks[0]
            if task_id not in first_task.dependencies:
                first_task.dependencies.append(task_id)

            # Add all loop tasks to workflow without modifying their dependencies further
            for task_obj in loop_tasks:
                self.workflow.add_task(task_obj)

        self._current_task = task_id
        return self

    def retry(
        self,
        max_retries: int = 3,
        delay: timedelta = timedelta(seconds=5),
        backoff_factor: float = 2.0,
    ) -> "WorkflowBuilder":
        if self._current_task and isinstance(
            self.workflow.tasks[self._current_task], TaskOperator,
        ):
            self.workflow.tasks[self._current_task].retry_policy = RetryPolicy(
                max_retries=max_retries, delay=delay, backoff_factor=backoff_factor,
            )
        return self

    def timeout(
        self, timeout: timedelta, kill_on_timeout: bool = True,
    ) -> "WorkflowBuilder":
        if self._current_task and isinstance(
            self.workflow.tasks[self._current_task], TaskOperator,
        ):
            self.workflow.tasks[self._current_task].timeout_policy = TimeoutPolicy(
                timeout=timeout, kill_on_timeout=kill_on_timeout,
            )
        return self

    # Phase 2: Event-based operators
    def emit_event(self, task_id: str, event_name: str, **kwargs: Any) -> "WorkflowBuilder":
        """Emit an event that other workflows can wait for."""
        task = EmitEventOperator(task_id=task_id, event_name=event_name, **kwargs)
        self._add_task(task, **kwargs)
        return self

    def wait_for_event(
        self, task_id: str, event_name: str, timeout_seconds: int | None = None, **kwargs: Any,
    ) -> "WorkflowBuilder":
        """Wait for an external event with optional timeout."""
        task = WaitForEventOperator(
            task_id=task_id, event_name=event_name, timeout_seconds=timeout_seconds, **kwargs,
        )
        self._add_task(task, **kwargs)
        return self

    # Phase 3: Callback hooks (applies to current task)
    def on_success(self, success_task_id: str) -> "WorkflowBuilder":
        """Set the task to run when the current task succeeds."""
        if self._current_task:
            self.workflow.tasks[self._current_task].on_success_task_id = success_task_id
        return self

    def on_failure(self, failure_task_id: str) -> "WorkflowBuilder":
        """Set the task to run when the current task fails."""
        if self._current_task:
            self.workflow.tasks[self._current_task].on_failure_task_id = failure_task_id
        return self

    # Phase 4: Switch operator
    def switch(
        self,
        task_id: str,
        switch_on: str,
        cases: dict[str, str],
        default: str | None = None,
        **kwargs: Any,
    ) -> "WorkflowBuilder":
        """Multi-branch switch/case operator."""
        task = SwitchOperator(
            task_id=task_id, switch_on=switch_on, cases=cases, default=default, **kwargs,
        )
        self._add_task(task, **kwargs)
        return self

    # Phase 1: Scheduling methods (delegate to Workflow)
    def set_schedule(self, cron: str) -> "WorkflowBuilder":
        """Set the cron schedule for this workflow."""
        self.workflow.set_schedule(cron)
        return self

    def set_start_date(self, start_date: datetime) -> "WorkflowBuilder":
        """Set when the schedule becomes active."""
        self.workflow.set_start_date(start_date)
        return self

    def set_catchup(self, enabled: bool) -> "WorkflowBuilder":
        """Set whether to backfill missed runs."""
        self.workflow.set_catchup(enabled)
        return self

    def set_paused(self, paused: bool) -> "WorkflowBuilder":
        """Set whether the workflow is paused."""
        self.workflow.set_paused(paused)
        return self

    def add_tags(self, *tags: str) -> "WorkflowBuilder":
        """Add tags to the workflow."""
        self.workflow.add_tags(*tags)
        return self

    def set_max_active_runs(self, count: int) -> "WorkflowBuilder":
        """Set maximum number of concurrent runs."""
        self.workflow.set_max_active_runs(count)
        return self

    def set_default_retry_policy(self, policy: RetryPolicy) -> "WorkflowBuilder":
        """Set default retry policy for all tasks."""
        self.workflow.set_default_retry_policy(policy)
        return self

    def build(self) -> Workflow:
        if not self.workflow.start_task and self.workflow.tasks:
            self.workflow.start_task = next(iter(self.workflow.tasks.keys()))
        return self.workflow
